local L = LibStub("AceLocale-3.0"):GetLocale("Grid2Options")
local BZ = LibStub("LibBabble-Zone-3.0"):GetUnstrictLookupTable()
local BB = LibStub("LibBabble-Boss-3.0"):GetUnstrictLookupTable()

local GSRD= Grid2:GetModule("Grid2RaidDebuffs")
local RDDB= {}

local moduleList
local moduleListAll
local selectedModule= ""
local selectedInstance= ""
local selectedDebuffs 
local optionModules
local optionInstances
local optionDebuffs
local optionsDebuffsCache= {}
local fmt= string.format

local ICON_SKULL= "Interface\\TargetingFrame\\UI-TargetingFrame-Skull"
local ICON_CHECKED = READY_CHECK_READY_TEXTURE
local ICON_UNCHECKED = READY_CHECK_NOT_READY_TEXTURE

-- Debuffs modules call this function to retrieve debuffs table
-- Not very clean but RaidDebuffsOptions dont provide any global 
-- object to put this method. 
function Grid2Options:GetRaidDebuffsTable()
	return RDDB
end

local function GetModules(all)
	if not moduleList then
		moduleList= {}
		local enabledModules= GSRD.db.profile.enabledModules or {}
		for name,_ in pairs(enabledModules) do
			moduleList[name]= L[name]
		end
	end
	if not moduleListAll then
		moduleListAll= {}
		for name,_ in pairs(RDDB) do
			moduleListAll[name]= L[name]
		end
	end
	return all and moduleListAll or moduleList
end

local function GetInstances(module)
	local values= {}
	if module and module~="" then
		local instances= RDDB[module]
		if instances then
			for name,_ in pairs(instances) do
				values[name]= BZ[name] or name
			end
		end
	end	
	return values
end

local function SetEnableDebuff(status,instance, spellId, value)
	local dbx= status.dbx
	if not dbx.debuffs then 
		dbx.debuffs= {}	
	end
	if not dbx.debuffs[instance] then
		dbx.debuffs[instance]= {}
	end
	local debuffs= dbx.debuffs[instance]
	if value then
		debuffs[#debuffs+1]= spellId
		selectedDebuffs[spellId]= #debuffs
	else
		table.remove(debuffs,selectedDebuffs[spellId])
		selectedDebuffs[spellId]= nil
	end
end

local function FormatDebuffName(spellName,spellId) 
	local icon= selectedDebuffs[spellId] and ICON_CHECKED or ICON_UNCHECKED
	return fmt("  |T%s:0|t%s", icon, spellName)
end

local function EnableInstanceAllDebuffs(status)
	local module= selectedModule
	local instance= selectedInstance
	local dbx= status.dbx
	if not dbx.debuffs then 
		dbx.debuffs= {}	
	end
	local debuffsall= RDDB[module][instance]
	local debuffs= {}
	for _,values in pairs(debuffsall) do
		for _,spellId in ipairs(values) do
			debuffs[#debuffs+1]= spellId
			selectedDebuffs[spellId]= #debuffs
		end
	end
	dbx.debuffs[instance]= debuffs
end

local function DisableInstanceAllDebuffs(status)
	local instance= selectedInstance
	local debuffs= status.dbx.debuffs
	if debuffs and debuffs[instance] then
		debuffs[instance]= nil
		selectedDebuffs= {}
	end
end

local function RefreshDebuffsOptions()
	local items= optionDebuffs.args
	for key,value in pairs(items) do
		local spellId=tonumber(key)
		if spellId then
			items[key].name= FormatDebuffName(value.desc,spellId)
		end
	end
end

local function EnableDisableModule(status, module, state)
	local dbx= status.dbx
	if not dbx.debuffs then 
		dbx.debuffs= {}	
	end
	local rddbx= GSRD.db.profile
	if not rddbx.enabledModules then
		rddbx.enabledModules= {}
	end
	if state then
		local instances= RDDB[module]
		for name,instance in pairs(instances) do
			local debuffs= {}
			for _,boss in pairs(instance) do
				for _,spellId in ipairs(boss) do
					debuffs[#debuffs+1]= spellId
				end
			end	
			dbx.debuffs[name]= debuffs
			local cacheKey= module..name
			if optionsDebuffsCache[cacheKey] then
				optionsDebuffsCache[cacheKey]= nil
			end	
		end
		rddbx.enabledModules[module]= true
		moduleList[module]= L[module]	
	else
		if dbx.debuffs[instance] then dbx.debuffs[instance]= nil end
		if rddbx.enabledModules[module] then rddbx.enabledModules[module]= nil end
		if moduleList[module] then moduleList[module]= nil end
	end
	selectedModule= ""
	optionModules.values= GetModules()
end

local function CreateStandardDebuff(bossName,spellId,spellName)
	local baseKey = fmt("%s>%s", string.match(bossName, "^(.-) .*$") or bossName, spellName):gsub("[ %.\"!']", "")
	if not Grid2:DbGetValue("statuses", baseKey) then
		-- Save status in database
		local dbx = {type = "debuff", spellName = spellId, color1 = {r=1, g=0, b=0, a=1} }
		Grid2:DbSetValue("statuses", baseKey, dbx) 
		--Create status in runtime
		local status = Grid2.setupFunc[dbx.type](baseKey, dbx)
		--Create the status options
		local funcMakeOptions = Grid2Options.typeMakeOptions[dbx.type]
		local optionParams = Grid2Options.optionParams[dbx.type]
		local options, subType = funcMakeOptions(Grid2Options, status, options, optionParams)
		Grid2Options:AddElementSubType("statuses", subType, status, options)
	end
end

local function MakeDebuffOptions(status,bossName,spellId,spellName,spellIcon)
	local options= {
		description={
			type="description",
			order= 10,
			name= spellName,
			fontSize= "large",
			image= spellIcon,
		},
		header1={
			type= "header",
			order= 20,
			name="",
		},
		enableSpell={
			type="toggle",
			order = 30,
			name = L["Enabled"],
			desc = "",
			get = function() return (selectedDebuffs[spellId] ~= nil) end,
			set = function(_, v)    
				SetEnableDebuff(status,selectedInstance,spellId,v)
				optionDebuffs.args[tostring(spellId)].name= FormatDebuffName(spellName,spellId)
			end,
		},	
		header2={
			type= "header",
			order= 40,
			name="",
		},
		createDebuff= {
			type = "execute",
			order = 50,
			name = L["Copy to Debuffs"],
			func = function() CreateStandardDebuff(bossName,spellId,spellName) end,
			disabled = false,
		}
	}
	return options
end

local function MakeDebuffsOptions(status)
	local module=selectedModule
	local instance=selectedInstance
	--
	selectedDebuffs= {}
	local dbx= status.dbx.debuffs[instance] or {}
	for index,value in ipairs(dbx) do
		selectedDebuffs[value]= index
	end
	--
	local cacheKey= module..instance
	local options= optionsDebuffsCache[cacheKey]
	if not options then
		options= {}
		options.enableall={
			type ="execute",
			order= 5,
			name = L["Enable All"],
			func= function() 
				EnableInstanceAllDebuffs(status)
				RefreshDebuffsOptions()
			end
		}
		options.disableall={
			type ="execute",
			order= 7,
			name = L["Disable All"],
			func= function() 
				DisableInstanceAllDebuffs(status)
				RefreshDebuffsOptions()
			end
		}
		local debuffs= RDDB[module][instance]
		local ORDER= 10
		for name,values in pairs(debuffs) do
			local bossName= BB[name] or name
			options[name]= {	
				type= "group",
				name=  fmt("|T%s:0|t%s", ICON_SKULL, bossName),
				order= ORDER,
				args= {},
			}
			ORDER= ORDER+1
			for _,spellId in ipairs(values) do
				local spellName,_, spellIcon = GetSpellInfo(spellId)
				options[tostring(spellId)]= {	
					type= "group",
					name= FormatDebuffName(spellName,spellId),
					desc= spellName,
					order= ORDER,
					args= MakeDebuffOptions(status,bossName,spellId,spellName,spellIcon)
				}
				ORDER=ORDER+1
			end
		end
		optionsDebuffsCache[cacheKey]= options
	end
	return options
end

local function MakeModulesListOptions(self,status,options,optionParams)
	options.header= {
		type="header",
		order= 149,
		name="",
	}
	options.modules= {
		type = "multiselect",
		name= L["Enabled raid debuffs modules"],
		order= 150,
		width= "full",
		get= function(info,key)
			return (moduleList[key] ~= nil)
		end,
		set= function(_,key,value)
			EnableDisableModule(status,key,value)
		end,
		values= GetModules(true)
	}
	return options
end

local function MakeGeneralOptions(self, status, options, optionParams)
	options= options or {}
	options = Grid2Options:MakeStatusStandardOptions(status, options, optionParams)
	options = Grid2Options:MakeStatusMissingOptions(status, options, optionParams)
	options = Grid2Options:MakeStatusBlinkThresholdOptions(status, options, optionParams)
	options = MakeModulesListOptions(self,status,options,optionParams)
	return options
end

local function MakeAdvancedOptions(self,status,options,optionPararms)
	optionModules= {
		type = "select",
		order = 10,
		name = L["Select module"],
		desc = "",
		get = function ()
			if selectedModule=="" then
				selectedModule= next(GetModules()) or ""
				selectedInstance= ""
				optionInstances.values= GetInstances(selectedModule)
				optionDebuffs.name= ""
				optionDebuffs.args= {}
			end
			return selectedModule
		end,
		set = function (info, v)
			selectedModule= v
			selectedInstance=""
			optionInstances.values= GetInstances(v)
			optionDebuffs.name= ""
			optionDebuffs.args= {}
		end,
		values= GetModules()
	}
	optionInstances= {
		type = "select",
		order = 20,
		name = L["Select instance"],
		desc = "",
		get = function ()
			return selectedInstance
		end,
		set = function (_, v)
			selectedInstance= v
			optionDebuffs.name= BZ[v] or v
			optionDebuffs.args= MakeDebuffsOptions(status)
		end,
		values= {}
	}
	optionDebuffs= {
		type="group",
		name="",
		order= 30,
		childGroups= "tree",
		args= {},
	}
	local options= {}
	options.modules  = optionModules
	options.instances= optionInstances
	options.debuffs  = optionDebuffs
	return options
end

local function MakeStatusOptions(self, status, options, optionParams)
	local generalOptions = MakeGeneralOptions(self,status,options,optionParams) 
	local advancedOptions= MakeAdvancedOptions(self,status)
	local options = {
		type = "group",
		childGroups= "tab",
		args= {
			general= {
				type= "group",
				name= L["General"],
				order= 10,
				args= generalOptions,
			},
			advanced= {
				type= "group",
				name= L["Advanced"],
				order= 20,
				args= advancedOptions,
			},
		},
	}
	return options
end

local prev_MakeStatusHandlers = Grid2Options.MakeStatusHandlers
function Grid2Options:MakeStatusHandlers(reset)
	prev_MakeStatusHandlers(self,reset)
	self:AddOptionHandler("raid-debuffs", MakeStatusOptions)
end